//
//	chroma noise reduction (version 2.2) - Avisynth filter reducing noise on chroma
//	Copyright (C) 2002 Marc Fauconneau
//
//	Inspired by :
//  chroma noise reduction (version 1.1) - VirtualDub filter reducing noise on chroma
//  Copyright (C) 2000 Gilles Mouchard
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  Please contact me for any bugs or questions.
//  marc.fd@libertysurf.fr

//  Change log :
//         01/08/2002 - ver 2.2  - Ugly bug fixed
//         31/07/2002 - ver 2.1  - Bug Fixes (thx to dividee ;)
//         20/07/2002 - ver 2.0  - Avisynth filter coded (from scratch)

#include "avisynth.h"
#include <math.h>

class Cnr2 : public GenericVideoFilter {
	int lastn;
	int lt[513];
	int ut[513];
	int vt[513];
	PVideoFrame pre;
	void CnrInit(int mode, int ln, int lm, int un, int um, int vn, int vm);
public:
	Cnr2(PClip _child, int mode, int ln, int lm, int un, int um, int vn, int vm) 
		: GenericVideoFilter(_child) { CnrInit(mode,ln,lm,un,um,vn,vm); }
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

void Cnr2::CnrInit(int mode, int ln, int lm, int un, int um, int vn, int vm) {
	__asm emms
	lastn = -2;
	if ( (mode<(-1)) || mode>(3) ) {lastn=(-666);}
	const double pi = 3.1415926535;
	for (int i=-256;i<256;i++) {lt[i+256] = 0;}
	for(int j = -ln; j <= ln; j++) {
		if (mode>=0) { lt[j+256] = int( lm / 2 * ( 1 + cos(j*j*pi/(ln*ln)))); 
		} else { lt[j+256] = int( lm / 2 * ( 1 + cos(j * pi / ln))); } 
	}
	for (i=-256;i<256;i++) {ut[i+256] = 0;}
	for(j = -un; j <= un; j++) {
		if (mode==1 || mode==3) { ut[j+256] = int( um / 2 * ( 1 + cos(j*j*pi/(un*un)))); 
		} else { ut[j+256] = int( um / 2 * ( 1 + cos(j * pi / un))); }
	}
	for (i=-256;i<256;i++) {vt[i+256] = 0;}
	for(j = -vn; j <= vn; j++) {
		if (mode==2 || mode==3) { vt[j+256] = int( vm / 2 * ( 1 + cos(j*j*pi/(vn*vn)))); 
		} else { vt[j+256] = int( vm / 2 * ( 1 + cos(j * pi / vn))); }
	}
}

PVideoFrame __stdcall Cnr2::GetFrame(int n, IScriptEnvironment* env) {
	if (lastn==(-666)) { env->ThrowError("Cnr2Ex : Mode should be -1,0,1,2,or 3"); }
	int mod_n = n;
	if (mod_n<0) {mod_n = 0;}

	PVideoFrame src = child->GetFrame(mod_n, env);

	if (lastn+1!=n) {
		pre = env->NewVideoFrame(vi);
		env->MakeWritable(&pre);
		pre = child->GetFrame(mod_n, env);
	}
	env->MakeWritable(&pre);

	const unsigned char *srcp = src->GetReadPtr();
	unsigned char *prep = pre->GetWritePtr();

	for (int fy = 0; fy < pre->GetHeight(); fy++) {
        for (int x = 0; x < pre->GetPitch(); x+=4) {

			int u = srcp[x+1]/*-0x80*/;
			int v = srcp[x+3]/*-0x80*/;

			const signed int ydiff = ((srcp[x+0]+srcp[x+2]) - (prep[x+0]+prep[x+0]))>>1;
			const signed int udiff = u - (prep[x+1]/*-0x80*/);
			const signed int vdiff = v - (prep[x+3]/*-0x80*/);
		
			const int chroma1_ratio = (lt[ydiff+256] * ut[udiff+256]);
			const int chroma2_ratio = (lt[ydiff+256] * vt[vdiff+256]);

			u = (chroma1_ratio * (prep[x+1]/*-0x80*/) + (65536 - chroma1_ratio) * u);
			v = (chroma2_ratio * (prep[x+3]/*-0x80*/) + (65536 - chroma2_ratio) * v);

			//u = (u >> 16) + ((u >> 15) & 1);
			u = (u + 0x8000) >> 16;
			//v = (v >> 16) + ((v >> 15) & 1);
			v = (v + 0x8000) >> 16;

			prep[x+0] = srcp[x+0];
			prep[x+2] = srcp[x+2];
			prep[x+1] = u/*+0x80*/;
			prep[x+3] = v/*+0x80*/;
		}
		srcp += src->GetPitch();
		prep += pre->GetPitch();
	}
	lastn = n;
	return pre;
}

AVSValue __cdecl Create_Cnr2(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return new Cnr2(args[0].AsClip(),args[1].AsInt(0),
					args[2].AsInt(35),args[3].AsInt(192),
					args[4].AsInt(47),args[5].AsInt(256),
					args[6].AsInt(47),args[7].AsInt(256));
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit(IScriptEnvironment* env) {
    env->AddFunction("cnr2", "c[mode]i[ln]i[lm]i[un]i[um]i[vn]i[vm]i", Create_Cnr2, 0);
    return "`cnr2' Chroma Noise Reducer 2.2";
}
